/*  algebra.h  */

#if !defined ALGEBRA_DEF
#define ALGEBRA_DEF

#include "ints.h"

/* the type algebra implements a sub(Hopf)algebra of the Steenrod algebra */

typedef struct {
  int8 is_restricted;     /* 0 => entire Steenrod algebra */
  uint16 badbits[15];     /* the forbidden bits           */
  uint16 topdeg;          /* dimension of the top class   */ 
} algebra;

void copy_algebra(algebra *a, algebra *b);
int algebras_are_equal(algebra *a, algebra *b);

int algebra_is_alg_n(algebra *a); 
/* returns -1 if a is none of the A(n), otherwise n */

char *algebra_name(algebra *a);
/* returns printable algebra description (profile or "A(n)") */

char *algebra_path_name(algebra *a);
/* returns path for the algebra resolution */

extern algebra thealgebra;   /* the one that is set */ 

void set_algebra_An(int n);     /* makes thealgebra = A(n) */
void set_algebra_Ainfty(void);  /* thealgebra = entire Steenrod algebra */

void clear_algebra(algebra *a);

#endif
