/*  ints.c  */

#include "ints.h"
#include <stdio.h>

int gis8(void) {
  int res = 0; int8 x = 1; 
  while (x) { res++, x <<= 1; }
  return res;
}

int gis16(void) {
  int res = 0; int16 x = 1; 
  while (x) { res++, x <<= 1; }
  return res;
}

int gis32(void) {
  int res = 0; int32 x = 1; 
  while (x) { res++, x <<= 1; }
  return res;
}

int gis64(void) {
  int res = 0; int64 x = 1; 
  while (x) { res++, x <<= 1; }
  return res;
}

int test_unit(void) { /* returns nonzero if something is wrong */
  unit u; int h;
  h = 0; u = 1;
  while (u) { h++, u <<= 1; }
  if ((1 << UNSIZE) != h) return 1;
  return 0;
}

int soi(unsigned long x) { /* soi = size of int */
  int res=0; 
  while (x) { res++, x >>= 1; }
  return res;
}

void available_sizes(void) {
  unsigned char xc; unsigned short xs; unsigned int xi; unsigned long xl; long a;

  xc = 0; xc = ~xc;
  xs = 0; xs = ~xs;
  xi = 0; xi = ~xi;
  xl = 0; xl = ~xl;

  printf("available sizes:\n");
  printf("  char  : %2d bits\n",soi(a=xc));
  printf("  short : %2d bits\n",soi(a=xs));
  printf("  int   : %2d bits\n",soi(a=xi));
  printf("  long  : %2d bits\n",soi(a=xl));
  printf("with gcc there is also the `long long' type.\n");
  printf("You might try this for int64.\n");  

}

void test_ints(void) {
  static int tested = 0;
  int nok=0;

  if (tested) return;    /* test only once */
  tested = 1;

  nok |= gis8()  < 8;
  nok |= gis16() < 16;
  nok |= gis32() < 32;
  nok |= gis64() < 64;
  nok |= test_unit();

  if (nok) { /* not ok */
    printf("\ntrouble with integer sizes detected\n");
    printf("the definitions in ints.h have to be adjusted\n");
    printf("recompile the package afterwards\n");
    printf("current sizes:\n");
    printf("  int8  : %2d bits\n",gis8());
    printf("  int16 : %2d bits\n",gis16());
    printf("  int32 : %2d bits\n",gis32());
    printf("  int64 : %2d bits\n",gis64());
    if (test_unit()) printf("definition of UNSIZE probably wrong\n");
    available_sizes();
    error("modify, recompile, and try again");
  }
}
