/*  ints.h   */

#if !defined INTS_DEF
#define INTS_DEF

#include "error.h"

/* 
   integer types with known (minimum) bitlength

   eventually, all critical parts of the program should use these types
*/

/* customize these typedefs, so that int<n> has (at least) n bits */

typedef char      int8;
typedef short     int16;
typedef int       int32;
typedef long long int64;

/* ALWAYS use unsigned types if bit operations (>> or <<) are needed */

typedef unsigned char      uint8;
typedef unsigned short     uint16;
typedef unsigned int       uint32;
typedef unsigned long long uint64;

/* the `unit' type is used for the linear algebra stuff */

typedef uint64 unit;

#define UNSIZE 6                      /* 2^6 = 64 bits per unit */
#define UNMASK ((1 << UNSIZE) - 1)    /* 2^UNMASK - 1           */

/* test routine, should be called by initialization routines */
void test_ints(void);            

#endif
