/* mkdots.c    reads file "pic", writes dots & lineXX       */

#include <stdio.h>
#include <stdlib.h>

#define MAXDIMENS 1500
int ioc[MAXDIMENS]; /* ioc = index of crowdiness (!) */

int maxn = 0; /* maximal top. dimension (n=t-s) */

typedef struct {
  int s,t,n;
  float x,y;
} dot;

typedef struct {
  int s,n1,n2,type;
} line;

void error(char *p) {
  printf("ERROR: %s\n",p);
  exit(1);
}

#define MAXDOTS 250000
#define MAXLINES 250000

dot  dotlist[MAXDOTS];
line linelist[MAXLINES];

int numdots = 0;
int numlines = 0;

int pl=-1; /* set this to a positive value to debug reading of "pic" */

int readline(FILE *fp, line *l,dot *d) {
  l->s = d->s; l->n1 = d->n;
  if (2 != fscanf(fp," h%d:%d ",&(l->type),&(l->n2))) return 0;
  if (pl-->0) printf("  line #%d h%d from [%d,%d]<--[%d,%d]\n",
							numlines,l->type,l->s,l->n1,l->s-1,l->n2);
  return 1;
}

int readdot(FILE *fp, dot *d) {
  if (3 == fscanf(fp," [%d,%d,%d] ",&(d->s),&(d->t),&(d->n))) {
	 d->x = (d->t) - (d->s);
	 d->y = (d->s);
	 if (d->x > maxn) 
		if ((maxn = d->x) >= MAXDIMENS) 
		  error("MAXDIMENS too small");
	 if (pl-->0) printf("  dot [s,n] = [%d,%d] (deg %d)\n",d->s,d->n,d->t);
	 return 1;
  }
  return 0;
}

int readloc(FILE *fp) {
  dot *d = dotlist+numdots; 
  line *l = linelist+numlines;
  if (MAXDOTS <= numdots) error("MAXDOTS too small");
  if (MAXLINES <= numlines) error("MAXLINES too small");
  if (!readdot(fp,d)) return 0;
  numdots++;
  for (;readline(fp,linelist+numlines,d);l++,numlines++) 
	 if (MAXLINES <= numlines) error("MAXLINES too small");
  return 1;
}

int comparedots(void *a, void *b) {
  dot *aa = (dot *) a;
  dot *bb = (dot *) b;
  int x;
  if ((x = (aa->s - bb->s))) return x;
  if ((x = (aa->n - bb->n))) return x;
  return a-b;
}

typedef int (*sf)(const void *, const void *);

void sortdots(void) {
  qsort(dotlist,numdots,sizeof(dot),(sf) comparedots);
}

int samelocdots(dot *a, dot *b) {
  return (a->s == b->s) && (a->t == b->t);  
}

void makeioc(void) {
  int n; 
  printf("checking crowdiness...\n");
  for (n=0;n<=maxn;n++) ioc[n] = 1;
  for (n=0;n<numdots;) {
	 int m,k,x; 
	 x = dotlist[n].t-dotlist[n].s;
	 m=n;
	 while ((m < numdots) && samelocdots(dotlist+n,dotlist+m)) m++;
	 k=m-n;
	 if (k > ioc[x]) ioc[x] = k;
	 n=m;
  }
}

/* we assume dotlist is sorted */
void movedots(void) {
  int n;
  makeioc();
  for (n=0;n<numdots;) {
	 int m,k;
	 m = n+1;
	 while ((m<numdots) && samelocdots(dotlist+n,dotlist+m)) m++;
	 /* dots n,n+1,...,m-1 have to be moved out */
	 k = m-n;
	 if (k>=1) {  /* k = number of dots in this bidegree */
		float x; 
		float y; 
		float dist; 
		float off; 
		int r; int i = ioc[dotlist[n].t-dotlist[n].s];  
		if (i>4) {   /* more than 4 dots = too crowded   */
		  if (k <= 4) {
			 i = 4;
		  } else {
			 i = k;
		  }
		}
		if (i>1) {
		  dist = 0.4;   
		  dist /= (i-1);
		  off = +0.2;
		  off -= (k-1) * dist;
		} else {
		  dist = 0;
		  off = 0;
		}
		x = dotlist[n].t - dotlist[n].s;
		y = dotlist[n].s;
		for (r=n;r<m;r++,off+=dist) {
		  dotlist[r].x = x + off;
		  dotlist[r].y = y - off*0.4;
		}
	 }
	 n = m;
  }
}

void writedots(void) {
  FILE *fp; int n;
  printf("   dots\n");
  fp = fopen("dots","wt");
  for (n=0;n<numdots;n++) 
	 fprintf(fp,"%.3f %.3f\n",dotlist[n].x,dotlist[n].y);
  fclose(fp);
}

int cdaux(void *a, void *b) {
  dot *aa = (dot *) a;
  dot *bb = (dot *) b;
  int x;
  if ((x = (aa->s - bb->s))) return x;
  if ((x = (aa->n - bb->n))) return x;
  return 0;
}

int theline;
dot *getdot(int s, int n) {
  dot aux, *res;
  aux.s = s; aux.n = n;
  res = (dot *) bsearch(&aux,dotlist,numdots,sizeof(dot),(sf) cdaux);
  if (NULL == res) {
	 int i,s2,n2;
	 printf("getdot error looking up s=%d,n=%d\nDotlist:\n",s,n);
	 printf("from line #%d = [%d,%d]-->[%d,%d] type %d\n",theline,
			  linelist[theline].s,
			  linelist[theline].n1,
			  linelist[theline].s-1,
			  linelist[theline].n2,
			  linelist[theline].type);
	 for (i=0;i<numdots;i++) {
		printf("dot #%3d = { s = %d, n = %d, t = %d, x = %.2f, y = %.2f }\n",
				 i,
				 s2 = dotlist[i].s,
				 n2 = dotlist[i].n,
				 dotlist[i].t,
				 dotlist[i].x,
				 dotlist[i].y);
		if (s2 > s) break;
		if ((s2 == s) && (n2 > n)) break;
	 }
	 printf("... (numdots = %d)\n",numdots);
	 return NULL;
	 exit(1);
  }
  return res;
}

void writehilines(int i) {
  FILE *fp; char fname[13]; int n;
  sprintf(fname,"lineh%d",i);
  fp = fopen(fname,"wt");
  printf("   %s\n",fname);
  for (n=0;n<numlines;n++)
	 if (linelist[n].type == i) {
		dot *sd,*td;
		theline=n; 
		sd = getdot(linelist[n].s,linelist[n].n1);
		td = getdot(linelist[n].s-1,linelist[n].n2);
		if ((NULL != sd) && (NULL != td)) 
		  fprintf(fp,"%.3f %.3f %.3f %.3f\n",sd->x,sd->y,td->x,td->y);
	 } 
  fclose(fp);
}

void writelines(void) {
  int n;
  for (n=0;n<8;n++) writehilines(n);
  
  writehilines(20);
  writehilines(21);
  
}

int main(void) {  
  FILE *fp;
  
  printf("reading pic..."); fflush(stdout);
  fp = fopen("pic","rt");
  while (readloc(fp)) ;  
  fclose(fp);
  printf(" ok.\n");
  
  sortdots();
  movedots();
  
  printf("numdots = %d, numlines = %d\n",numdots,numlines); 
  
  printf("writing files:\n");
  writedots();
  writelines();
  printf("finished.\n");

  return 0;
}

