/* modules.c   -     elements, modules, etc.  */

#include "modules.h"

#define MODULE_VERSION 4    /* version number of the code that follows */

int doc = 0;

void init_stoplist_buffer(void) { 
  printf("dynamic stoplist memory 11/4/98\n"); 
}

int malloc_stoplist(stoplist *s, int num) {
  s->num = num;
  s->dat = malloc(sizeof(stop)*(1+num));
  if (NULL == s->dat) error("memory exhausted in malloc_stoplist()");
  return 1;
}

void dispose_stoplist(stoplist *s) {
  if (NULL != s->dat) free(s->dat);
  s->dat = NULL;
}

void fprint_stoplist(FILE *fp, stoplist *s) {
  int n; stop *q = s->dat;
  for (n=0;n<s->num;n++,q++) {
    fprintf(fp," ");
    fprintf_stop(fp,q);
  }
  printf(" <end>\n");
}

void print_stoplist(stoplist *s) {
  fprint_stoplist(stdout,s);
}

/* the following strange construct seems to be the unique solution */
/* to the problem of updating the header on my system.             */
void update_header(module_header *hd) {
  char fnme[50]; FILE *fp; int a;
  sprintf(fnme,"%s/C%d",algebra_path_name(&(hd->data.alg)),hd->data.sdeg);
  sprintf(fnme,"%s",hd->data.fnme);
  fp=fopen(fnme,"r+b");
  if (NULL != fp) {
	 a = fgetc(fp); ungetc(a,fp);
  }
  if ((NULL == fp) || feof(fp) || ferror(fp)) {
    if (NULL != fp) fclose(fp);
    fp = fopen(fnme,"wb");
  }
  rewind(fp);
  fseek(fp,0L,SEEK_SET);
  hd->data.up_date = time(NULL);
  if (1 != fwrite(hd,sizeof(module_header),1,fp)) {
    printf("Trouble in update_header (file=%s):\n",fnme);
    if (feof(fp)) printf("feof ");
    if (ferror(fp)) printf("ferror ");
    perror("system error string");
    error("fwrite failed in update_header");
  }
  fclose(fp);
}

void header_update_3_to_4(module_header *hd) {
  if (hd->data.version == 3) {
    clear_algebra(&(hd->data.alg));
    hd->data.version = 4;
  }
}

int read_header(FILE *fp, module_header *hd) {
  if (1 != fread(hd,sizeof(module_header),1,fp)) return 0;
  if (26219 != hd->data.mpn) {
    print_header(hd);
    error("data file corrupted, mpn incorrect");
  }
  if (3 == hd->data.version) header_update_3_to_4(hd);
  if (MODULE_VERSION != hd->data.version) {
    printf("datafile has format #%d, program format #%d\n",hd->data.version,MODULE_VERSION);
    error("different data format");
  }
  return 1;
}

char *tmstr(time_t t, char *s) {
  struct tm *tmp = localtime(&t);
  sprintf(s,"%d/%d/%02d %2d:%02d",tmp->tm_mon+1,tmp->tm_mday,tmp->tm_year%100,tmp->tm_hour,tmp->tm_min);
  return s;
}

void print_header(module_header *hd) {
  char s1[50],s2[50];
  printf("%s, created %s, updated %s, complete to %d\n"
	 ,hd->data.fnme,
	 tmstr(hd->data.cr_date,s1),tmstr(hd->data.up_date,s2),
	 hd->data.tmax);
}

void print_sq0header(module_header *hd) {
  char s1[50],s2[50];
  printf("doubl.%d, created %s, last update %s, complete to %d\n"
	 ,hd->data.sdeg,
	 tmstr(hd->data.cr_date,s1),tmstr(hd->data.up_date,s2),
	 hd->data.tmax);
}

int read_gen_aux(FILE *fp, gen *g) {
   int a[3];
   if (2 != fread(a,sizeof(int),2,fp)) {
      if (feof(fp)) return 0;
      error("read_gen_aux");
   }
   g->num = a[0];
   g->deg = a[1];
   return 1;
}

void write_gen_aux(FILE *fp, gen *g) {
   int a[3];
   a[0] = g->num;
   a[1] = g->deg;
   if (2 != fwrite(a,sizeof(int),2,fp)) 
      error("write_gen_aux(...)");
}

int read_gen(FILE *fp,gen *g, module *m, genfunc gf, stopfunc sf, genfunc ef, int mode) {
   stop aux, *s=&aux; int num=0;
   if (read_gen_aux(fp,g)) {
     if (g->deg > m->header.data.tmax) {
		 printf("read_gen(...): gen of degree %d found in file, "
				  "though tmax = %d.\n",g->deg,m->header.data.tmax);
	  } 
     if (NULL != gf) if (0 == gf(m,m->numgens)) return 0;
     if (0 == (mode & MOD_NOLOAD)) {
       malloc_stoplist(&(g->diff),g->num);
       s = g->diff.dat;
     }
     while (read_stop(fp,s)) { 
       num++;
       if (NULL != sf) if (0 == sf(m,m->numgens,s)) return 0;
       if (0 == (mode & MOD_NOLOAD)) s++;
     }
     if (num != g->num) {
       printf("\nread_gen: %d stops read, %d (g->num) stops announced\n",num,g->num);
       error("inconsistency detected by read_gen()");
     }
     if (NULL != ef) if (0 == ef(m,m->numgens)) return 0;
     return 1;
   }
   return 0;
}

int hf_print(module *m) {
  print_header(&(m->header)); return 1;
}

int hf_sq0print(module *m) {
  print_sq0header(&(m->header)); return 1;
}


int gf_print(module *m, int n) {
  printf("\n<%d,%d> (%d summands):",m->gens[n].deg,n,m->gens[n].num);
  return 1;
}

int sf_print(module *m, int n, stop *s) {
  printf(" "); fprintf_stop(stdout,s); return 1;
}

int sf_invert(module *m, int n, stop *s) {
  invert_stop(s); return 1;
}

int read_module(module *m, int s) {
  if (s<1) error("read_module(m,s) with s < 1");
  return parse_modulefile(s,m,hf_print,NULL,NULL,NULL,MOD_NORMAL);
}

int read_sq0file(module *m, int s) {
  char fnme[50];
  if (s<1) error("read_sq0file(m,s) with s < 1");
  sprintf(fnme,"data.sq0/doubl.%d",s);
  return parse_datafile(fnme,s,m,hf_sq0print,NULL,NULL,NULL,MOD_NORMAL);
}
  
int read_hurfile(module *m, int s, int n) {
  char fnme[50];
  if (s<1) error("read_hurfile(m,s,n) with s < 1");
  sprintf(fnme,"data.hur%d/hur.%d",n,s);
  return parse_datafile(fnme,s,m,hf_print,NULL,NULL,NULL,MOD_NORMAL);
}
  
int read_module_no_diffs(module *m, int s) {
  if (s<1) error("read_module_no_diffs(m,s) with s < 1");
  return parse_modulefile(s,m,hf_print,NULL,NULL,NULL,MOD_NOLOAD);
}

int parse_modulefile(int s, module *m, headfunc hf, genfunc gf, stopfunc sf, genfunc ef, int mode) {
  char fnme[50];
  sprintf(fnme,"%s/C%d",algebra_path_name(&(thealgebra)),s);
  return parse_datafile(fnme,s,m,hf,gf,sf,ef,mode);
}

int parse_datafile(char *fnme, int s, module *m, headfunc hf, genfunc gf, stopfunc sf, genfunc ef, int mode) {
   FILE *fp; int a; int newfile=0;
	sprintf(m->header.data.fnme,"%s",fnme);
   m->sdeg = s; m->numgens = 0;
   newfile = (NULL == (fp = fopen(fnme,"rb")));
   if (!newfile) {
	  a = fgetc(fp); ungetc(a,fp);
	  newfile |= ferror(fp) || feof(fp);
   }
	if (newfile) {
     /* new module */
     if (NULL != fp) fclose(fp);
     if (mode & MOD_NOCREATE) return 0;
     m->header.data.cr_date = time(NULL);
     m->header.data.mpn = 26219;
     m->header.data.version = MODULE_VERSION;
     m->header.data.sdeg = s;
     m->header.data.tmax = 0; /* was s-1 */
     copy_algebra(&(m->header.data.alg),&thealgebra);
     update_header(&(m->header));
     if (NULL == (fp = fopen(fnme,"rb"))) 
       error("parse_modulefile: this shouldnt happen");
   } 
   if (!read_header(fp,&(m->header)))
     error("parse_module: this shouldnt happen (II)");
	sprintf(m->header.data.fnme,"%s",fnme);
   if (NULL != hf) if (0 == hf(m)) return 1;
   if (!(mode & MOD_HEADER)) 
	  while (read_gen(fp,&(m->gens[m->numgens]),m,gf,sf,ef,mode)) 
		 if (++m->numgens == MAXGENS) error("MAXGENS too small");
   clearerr(fp);
   fclose(fp);
   return 1;
}


void write_gen(FILE *fp, gen *g) {
   stop *s; int n;
   if (g->num != g->diff.num) 
     error("inconsistency (=improper use) detected in write_gen()");
   write_gen_aux(fp,g);
   for (n=0,s=g->diff.dat;n<g->num;n++,s++) write_stop(fp,s);
   write_enddiff(fp);
}

void dispose_module(module *m) {
  int n;
  for (n=m->numgens;n--;) dispose_stoplist(&(m->gens[n].diff));
  m->numgens = 0;
}

typedef int (*sf)(const void *, const void *);
   /* sf = search function */

void sort_element_same_sig(element *e) { 
   qsort(e->dat,e->num,sizeof(stop),(sf) compare_stops_same_signature);
#ifdef USEMMX
	MMXCLEAR;
#endif
}

void sort_element(element *e) { 
   qsort(e->dat,e->num,sizeof(stop),(sf) compare_stops);
#ifdef USEMMX
	MMXCLEAR;
#endif
}

void modulebasis(element *b, module *m, int t, signature *sig) {
   int n;
   b->num = 0;
   for (n=0;n<m->numgens;n++) {
      int dg = m->gens[n].deg; int k;
      if (dg > t) break;
      dg = t-dg;
      k = makebasis(dg,sig,n,&(b->dat[b->num]));
      b->num += k;
      if (b->num > MAXELEMENT) error("MAXELEMENT too small");
   }
   sort_element_same_sig(b);
}

int countmodulebasis(module *m, int t, int max) {
   int n; stop sig; int res=0;
	clear_stop(&sig);
   for (n=0;n<m->numgens;n++) {
      int dg = m->gens[n].deg; 
      if (dg > t) break;
      dg = t-dg;
      res +=count(dg,&sig);
		if (res > max) break;
   }
	return res;
}

int num_stop(element *e, stop *s) {
   stop *r;
   r = bsearch(s,e->dat,e->num,sizeof(stop),(sf) compare_stops);
   if (NULL == r) return -1; /* not found */
   return r - e->dat;
}

int num_stop_same_signature(element *e, stop *s) {
   stop *r;
   r = bsearch(s,e->dat,e->num,sizeof(stop),(sf) compare_stops_same_signature);
   if (NULL == r) return -1; /* not found */
   return r - e->dat;
}

void add_stop_to_element(element *e, stop *s) {
  int i,r=num_stop(e,s);
  if (-1 == r) {
	 if (e->num >= MAXELEMENT) error("MAXELEMENT too small");
	 copy_stop(&(e->dat[e->num]),s);
	 e->num++;
	 sort_element(e);
	 return;
  }
  e->num--;
  for (i=r;i<e->num;i++) copy_stop(&(e->dat[i]),&(e->dat[i+1]));
}


void make_zero_module(module *m) {
   m->numgens = 0;
   m->header.data.tmax = 11111;
}

void make_C0_module(module *m) {
   m->numgens = 1; m->sdeg = 0; 
   m->gens[0].deg = 0;
   m->gens[0].diff.num=0;
   m->gens[0].num = 0;
   m->header.data.tmax = 11111;
}

void fake_doubl0(module *m) {
   m->numgens = 1; m->sdeg = 0; 
   m->gens[0].deg = 0;
	malloc_stoplist(&(m->gens[0].diff),1);
	clear_stop(&(m->gens[0].diff.dat[0]));
	m->gens[0].diff.dat[0].id = 0;
   m->gens[0].num = 0;
   m->header.data.tmax = 11111;
}

void fake_hur0(module *m) { 
  fake_doubl0(m);
}

void fprint_element(FILE *fp, element *e) {
   int n;
   fprintf(fp,"element %d summands:\n",e->num);
   for (n=0;n<e->num;n++) {
      fprintf(fp," ");
      fprintf_stop(fp,&(e->dat[n]));
   }
   fprintf(fp,"\n");
}

void print_element(element *e) {
    fprint_element(stdout, e);
}

int negstopmode=0;

void dump_module_basic(FILE *fp, module *m, int type) {
  int i;
  switch (type) {
  case DUMP_ASCII: 
  case DUMP_TEX:
	 break;
  default:
	 error("dump_module(...): wrong type");
  }
  switch (type) {
  case DUMP_ASCII:
	 fprintf(fp,"Module: filename=%s number of gens=%d\n",
				m->header.data.fnme,m->numgens);
	 break;
  case DUMP_TEX:
	 fprintf(fp,"\n\n"
				   "Module {\\tt %s}\\\\\n %d generator%c\\\\\n", 
				m->header.data.fnme,m->numgens,(m->numgens != 1) ? 's' : ' ');
	 break;
  default:
	 error("dump_module(...): wrong type");  
  }
  for (i=0;i<m->numgens;i++) {
	 switch (type) {
	 case DUMP_ASCII:
		fprintf(fp,"[%d] %d: ",i,m->gens[i].deg);
		fprint_stoplist(fp,&(m->gens[i].diff));
		break;
	 case DUMP_TEX:
		fprintf(fp,"\\begin{tabular}{p{3mm}p{5mm}l}\n\\hline\n"
				  "%d & [%d] & \\begin{minipage}{11cm}\n",m->gens[i].deg,i);
		fprint_stoplist(fp,&(m->gens[i].diff));
		fprintf(fp,"\n\\end{minipage}\n\\end{tabular} \\\\\n");
		break;
	 default:
		error("dump_module(...): wrong type");  
	 }
  }
  switch (type) {
  case DUMP_ASCII:
	 break;
  case DUMP_TEX:
	 fprintf(fp,"\n");
	 break;
  default:
	 error("dump_module(...): wrong type");
  }  
}

void dump_module(FILE *fp, module *m, int type) {
  negstopmode = 0;
  dump_module_basic(fp,m,type);
}

void gen_from_element(gen *g, element *e, int dg) {
  int i;
  malloc_stoplist(&(g->diff),e->num);
  for (i=0;i<e->num;i++) copy_stop(&(g->diff.dat[i]),&(e->dat[i]));
  g->num = e->num;
  g->deg = dg;
}

