/* profile.c   -   profiles and signatures */

#include "profile.h"

void print_profile(profile *pro) {
  int n;
  printf("profile information:\n");
  printf("  mask: ");
  for (n=0;n<NALG;n++) printf(" %d",pro->mask[n]);
  printf("\n  shft: ");
  for (n=0;n<NALG;n++) printf(" %d",pro->shft[n]);
  printf("\n  xdegrees: ");
  for (n=0;n<NALG;n++) printf(" %d",pro->xdegrees[n]);
  printf("\n  numpts = %d\n",pro->numpts);
  printf("  location of the bits (most significant first), (t,2^s) for P_t^s\n");
  for (n=0;n<pro->numpts;n++) {
    printf("    (%d,%d)\n",pro->datt[n]+1,pro->dats[n]);    
  }
  printf("  numxi = %d\n",pro->numxi);
}

void fppf(FILE *fp, profile *p) {
  int i; char c=' '; int aux;
  for (i=0;i<p->numxi;i++,c='|') {
    aux = p->mask[i];
    if (0 == (aux & (aux+1))) { /* aux+1 is a power of 2 */
      int n=0; while (0 != aux) { aux >>= 1; n++; }
      fprintf(fp,"%c%d",c,n);
    } else {
      fprintf(fp,"%c%d*",c,p->mask[i]);
    }
  }
}

void first_signature(signature *sig) {
  clear_stop(sig);
}

int next_signature(signature *sig) {
  int n = theprofile.numpts - 1,x;
  if (0 == theprofile.numpts) return 0;
  do {
    x = sig->dat[theprofile.datt[n]] & theprofile.dats[n];
    sig->dat[theprofile.datt[n]] ^= theprofile.dats[n];
  } while (n-- && x);
  if (x) return 0;
  return 1;
}

profile theprofile;  /* the profile that is being used */

/* from here on: routines that set/modify theprofile */

void theprofile_make_shft_xdegrees(void) {
  /* compute shft and xdegrees from the other values */
  int n,r,x;

  /* shft[n] is the largest s such that (1 << s) - 1 is a */
  /* "subset" of mask[n]                                  */
  /* xdegrees[n] = degrees[n] << shft[n]                  */

  for (n=0;n<theprofile.numxi;n++) {
    for (x=theprofile.mask[n],r=0;(1 & x);r++,x>>=1) ;
    theprofile.shft[n] = r;
    theprofile.xdegrees[n] = degrees[n] << r;
  }

}

int MAX(int x,int y) {
  return (x > y) ? x : y;
}

#define PWU error("(internal, profile.c) add_P(i,j): wrong usage");

void add_P(int i, int j) { /* add P_i^j to theprofile */
  int k;
  if (0 >= i) PWU; /* wrong usage */

  /* if (i+j>=7) return; */ /* avoid big profile (throw this out!!!!) */

  /* reject unallowable P_i^js */
  for (k=1;k<=j;k++) 
    if (0 == (theprofile.mask[i-1+k] & ((1 << j) >> k))) return;
  
  /* reject unreachable P_i^js */
  if (0 != ((thealgebra.badbits[i-1] | theprofile.mask[i-1]) & (1 << j)))
    return;

  theprofile.numxi = MAX(theprofile.numxi,i);
  theprofile.mask[i-1] |= 1 << j;
  theprofile.datt[theprofile.numpts] = i-1;
  theprofile.dats[theprofile.numpts] = 1 << j;
  theprofile.numpts++;
  
  theprofile_make_shft_xdegrees();
}

void clear_profile(void) {
  int n;
  theprofile.numxi=0;
  theprofile.numpts=0;
  for (n=0;n<NALG;n++) theprofile.mask[n] = 0;
  theprofile_make_shft_xdegrees();
}

void set_empty_profile(void) {
  clear_profile();
}

int deg_P(int i, int j) { /* degree of P_i^j */
  return ((1 << i) - 1) << j;
}

/* 
   max profile lower chooses B=theprofile such that Ext_B(F2) lies 
   above (s,t-correction).   
*/
void set_max_profile_lower(int s, int t) {
  int d,q,i,j,ok;   /* P_{i+1}^j , q = max. slope, d = max. dim */
  set_empty_profile();
  if ((t-s) <= 1) return; 
  i=0,j=0,q=1,d=1;
  clear_profile();
  do {
    add_P(i+1,j);
    if (i) { i--; j++; } else { i = j+1; j = 0; }
    
    q = MAX(q,(2 << i) - 1); 
    d += ((2 << i) - 1) << j;
    
    ok = t > (d+q*s);
  } while (ok);
  
  theprofile_make_shft_xdegrees();
}

/* 
   max profile lower chooses B=theprofile such that Ext_B(F2) lies 
   below (s,t).  (no correction to (s,t) necessary.) 
*/
void set_max_profile_upper(int s, int t) {
  int i,j,k,m;
  clear_profile();
  if (t-s <= 1) return;

  if (0 == s) return; /* use lower profile instead */

  k = 0; m = 1;
  while (t >= m*s) { m = (m << 1) + 1; k++; }
  
  for (j=0;j<NALG;j++)
    for (i=k;i+j<NALG;i++) 
      if (deg_P(i+1,j) <= (t-s)) add_P(i+1,j);

  /* 
     two versions are possible:
       normal    :  if (k--)        
       desperate :  while (k--)    [ not fully understood yet ] 

       desperate doesnt work !!
  */

  if (k--) 
    for (j=NALG;j--;) 
      if (deg_P(k+1,j) <= (t-s))  
	if (deg_P(k+1,j) * s > t) add_P(k+1,j); 
  
  theprofile_make_shft_xdegrees();
}
