/* seqno.c   --   sequence number of stops  */

#include "seqno.h"
#include "bits.h"
#include "algebra.h"

/*
  dim_table[k][n] = dimension of the polynomial algebra F2[xi_1,...,xi_k+1]
                    in degree n, with restrictions from thealgebra
*/

#define MAXSEQNO 1200 /* ambitious, eh? */

int dim_table[NALG+1][MAXSEQNO];  
int seq_table[NALG+1][MAXSEQNO];  /* used by seqno  */


typedef int (*seq_func) (stop *, short);
seq_func theseqfunc;

/* the following routine initializes (=computes) the tables we need */
void init_dimtab(int maxdeg) {
  int i,k,n;

  if (maxdeg < MAXSEQNO-10) maxdeg = MAXSEQNO-10;

  if (maxdeg++ > MAXSEQNO) error("init_dimtab(maxdegree): maxdegree too big\n");
  
  for (n=0;n<maxdeg;n++) 
    dim_table[0][n] = (0 == (n & thealgebra.badbits[0])) ? 1 : 0;

  for (k=1;k<=NALG;k++) 
    for (n=0;n<maxdeg;n++) {
      int res=0, d=degrees[k];
      /* compute dim_table[k][n] */
      for (i=n/d;i>=0;i--)
	if (0 == (i & thealgebra.badbits[k])) res += dim_table[k-1][n-i*d];
      dim_table[k][n] = res;
    }

  if (!thealgebra.is_restricted) {
	 for (k=1;k<NALG;k++)
		for (n=0;n<MAXSEQNO;n++) { 
		  int res=0, d = degrees[k];
		  /* compute seq_table[k][n] */
		  for (i=n-d;i>=0;i-=d) 
			 res += dim_table[k-1][i];
		  seq_table[k][n] = res;
		}
	 theseqfunc = opt_seqno_with_degree;
  } else theseqfunc = safe_seqno_with_degree;

/* seq_tab[k+1][n] = dim_tab[k][n-d] + dim_tab[k][n-2d] +...+ dim_tab[k][n%d] */

}

inline
int dim(int n) {     /* dimension of the algebra in degree n */
  return dim_table[NALG][n];
}


inline
int seqno(stop *s) { return seqno_with_degree(s,degree_stop(s,degrees)); }

/* the following routine does not work with restricted subalgebras */
#ifdef USEMMX
/* will be called within the "getloop", so must preserve mm5, mm6 and mm7 */
unit SQNDATL = 0x000f000700030001LL;
unit SQNDATH = 0x00ff007f003f001fLL;
unit sm0,sm1,sm2;
void pmmx(void) {
  asm volatile("movq %mm0,sm0\n\tmovq %mm1,sm1\n\tmovq %mm2,sm2\n\t");
  printf("mm0 = %Lx, mm1 = %Lx, mm2 = %Lx\n",sm0,sm1,sm2);
}
#define MM /*"call pmmx\n\t"*/
inline
int opt_seqno_with_degree(stop *s, short dg) { /* return sequence number of *s */
  int res=0; unsigned char k; 
  unsigned short products[8]; /* will become s->dat[..] * degrees[..] */
  asm volatile("movq (%0),%%mm0\n\t"
					"movq %%mm0,%%mm1\n\t"
					"pxor %%mm2,%%mm2\n\t" 
MM
					"punpcklbw %%mm2,%%mm0\n\t" /* mm0 = lower 4 words of s  */
					"punpckhbw %%mm2,%%mm1\n\t" /* mm1 = higher 4 words of s */
MM
					"pmullw SQNDATL,%%mm0\n\t"
					"pmullw SQNDATH,%%mm1\n\t"
MM
					: : "r" (s->dat));
  asm volatile("movq %%mm0,(%0)\n\t"
					"movq %%mm1,(%1)\n\t"
					: : "r" (&(products[0])), "r" (&(products[4])));
  /*
	 for (k=0;k<NALG;k++) 
	 printf("[%d * %d -> %d]\n",s->dat[k],degrees[k],products[k]);
	 for (k=0;k<NALG;k++) if (s->dat[k]*degrees[k] != products[k]) error("!!!");
  */

  for (k=NALG;--k;) {
	 res += seq_table[k][dg -= products[k]]; 
  }
  return res;
}
#else
inline
int opt_seqno_with_degree(stop *s, short dg) { /* return sequence number of *s */
  int res=0; unsigned char k;
  for (k=NALG;--k;) {
	 res += seq_table[k][dg -= (s->dat[k])*degrees[k]]; 
  }
  return res;
}
#endif

inline
int seqno_with_degree(stop *s, short dg) {
#ifdef NOSUBALG
  return opt_seqno_with_degree(s,dg);
#else
  return (theseqfunc)(s,dg);
#endif
}

int safe_seqno_with_degree(stop *s, short deg) { 
  /* return sequence number of *s */
  int16 k; int res=0;
  for (k=NALG;--k;) {
    int16 d=degrees[k], i=deg/d, e=s->dat[k];
    for (;i>e;i--) 
      if (0 == (i & thealgebra.badbits[k])) 
		  res += dim_table[k-1][deg-i*d];
    deg -= i*d;
  }
  return res;
}

#include "error.h"
#include "profile.h"
#include "stoputils.h"

#define MAXTD 90000

void test_seqno(int dg) {  
  char tbl[MAXTD];
  unsigned int i,cnt,sqn,max,err;
  stop s;

  printf(" (%d)",dg);
  fflush(stdout);

  if (dim(dg) >= MAXTD) error("test_seqno, err #1");

  for (i=0;i<MAXTD;i++) tbl[i]=0;

  set_empty_profile();

  err = cnt = 0; max = -1;

  if (first_stop(&s,degrees,dg)) 
    do {
      cnt++;
      sqn = seqno_with_degree(&s,dg);
		if (sqn < MAXTD) tbl[sqn]++;
      if (sqn > max) max = sqn;

      if (sqn != cnt-1) {
		  fprintf_stop(stdout,&s);
		  printf("  %d\n",sqn);
		  err=1;
      }

    } while (next_stop(&s,degrees));

  /*for (i=0;i<=max;i++) 
    if (1 != tbl[i]) { err=1; printf("tbl[%d]=%d\n",i,tbl[i]); };
  */
  if (err || (cnt != dim(dg))) {
    printf("dim(%d)=%d, count gives %d\n",dg,dim(dg),cnt);
    err = 1;
  }
  
  if (err) error("test_seqno: error detected");

}

void test_seqno_to(n) {
  int k;
  for (k=0;k<=n;k++) test_seqno(k);
  printf("\n");
}

void test_main(void) {

  initialize_stops();
  set_algebra_Ainfty();
  init_dimtab(500);

  printf("testing %s\n",algebra_name(&thealgebra));

  test_seqno_to(250);
  /*  
  for (n=0;n<=6;n++) {
    set_algebra_An(n);
	 init_dimtab(500);
    printf("testing %s\n",algebra_name(&thealgebra));
    test_seqno_to(250);
  }
  */
}
