/* stoputils.c   -   combined stuff for stop, profile, algebra */

#include "stoputils.h"
#include "error.h"
#include "bits.h"

int compare_stops_same_signature(stop *s, stop *t) {
  int x;
  x = s->id-t->id; if (0 != x) return x;
#ifdef USEMMXALWAYS  /* the non-MMX routine is efficient enough */
  /* this will be called within the "getloop" so mm5, mm6 and mm7 must
	  not be clobbered. In particular, "emms" is not allowed.           */
  asm volatile ("movq (%1),%%mm0\n\t"
					 "psubw (%2),%%mm0\n\t"
					 "movd %%mm0,%0\n\t"
					 : "=r" (x) : "r" (s->dat) , "r" (t->dat));
  if (x) return x;
  asm volatile ("psrlq $32,%%mm0\n\t"
					 "movd %%mm0,%0\n\t"
					 : "=r" (x));
  return x;
#else
	 /*
	 {
	 char n;
	 for (n=NALG;n--;) if (0 != (x = (s->dat[n]-t->dat[n]))) return x;
	 }
	 */
  { 
	 x = *((int *) &(s->dat[0])) -  *((int *) &(t->dat[0]));
	 if (0 != (int) x) return (int) x;
	 x = *((int *) &(s->dat[4])) -  *((int *) &(t->dat[4]));
	 return (int) (x);
  }
  return 0;
#endif
}

int compare_stops(stop *s, stop *t) {
  int n=0;
  for (;n<theprofile.numpts;n++) {
    int x = (s->dat[theprofile.datt[n]] & theprofile.dats[n])
      - (t->dat[theprofile.datt[n]] & theprofile.dats[n]);
    if (0 != x) return x;
  }
  return compare_stops_same_signature(s,t);
}

int stop_has_signature(stop *s, signature *sig) {
  int n=0;
  for (;n<NALG;n++)
    if ((s->dat[n] & theprofile.mask[n]) != sig->dat[n]) return 0;
  return 1;
}

void get_signature(signature *sig, stop *s) {
  int n=NALG;
  while (n--) sig->dat[n] = s->dat[n] & theprofile.mask[n];
}

void set_signature(stop *s, signature *sig) {
  int n=NALG;
  while (n--) 
    s->dat[n] = sig->dat[n] | 
      (s->dat[n] & (s->dat[n] ^ theprofile.mask[n]));
}

int xnext_stop(stop *s, int *deglist, int rem) {
  uint16 n,x; uint16 aux;
  do {
    rem += s->dat[0] * deglist[0]; 
    for (n=1;(n<NALG) && (0 == s->dat[n]);n++) ;
    if (n == NALG) return 0; 
    x = s->dat[n]-1;
    aux = theprofile.mask[n] | thealgebra.badbits[n];
    REMOVE_BADBITS(x,aux);  /* kill offending bits */
    rem += (s->dat[n] - x) * deglist[n]; s->dat[n] = x;
    while (n--) {
      x = rem / deglist[n];
      aux = theprofile.mask[n] | thealgebra.badbits[n];
      REMOVE_BADBITS(x,aux);  /* kill offending bits */
      s->dat[n] = x; rem -= x*deglist[n];
    }
  } while (rem != 0);
  return 1;
}

int first_stop(stop *s, int *deglist, int deg) { 
  uint16 n; uint16 aux;
  if (deg>255) return 0;
  for (n=NALG;n--;) {
    uint16 a = deg / deglist[n];
    aux = theprofile.mask[n] | thealgebra.badbits[n];
    REMOVE_BADBITS(a,aux);  /* kill offending bits */
    s->dat[n] = a; deg -= a*deglist[n];
  }
  if (0 == deg) return 1;
  return xnext_stop(s,deglist,deg);
}

int next_stop(stop *s, int *deglist) {
  return xnext_stop(s,deglist,0);
}

int makebasis(int deg, signature *sig, int id, stop *bas) {
  /* write stops of given degree and signature into the list bas. */
  /* stops get id "id". return number of basis elements.          */
  int num=0; stop aux; int n; 
  if (deg > MAXDEGREE) error("MAXDEGREE too small");
  clear_stop(&aux); deg -= degree_stop(sig,degrees); 
  if (deg < 0) return 0;
  
  if (first_stop(&aux,degrees,deg)) {  
    do {
      for (n=0;n<NALG;n++) 
		  bas->dat[n] = (aux.dat[n]) | sig->dat[n];
      bas->id = id; 
      bas++; num++; 
    } while (next_stop(&aux,degrees));
  }

  return num;
}

int count(int deg, signature *sig) {
  /* returns number of basis elements of given signature.         */
  int num=0; stop aux;  
  if (deg > MAXDEGREE) error("MAXDEGREE too small");
  clear_stop(&aux); deg -= degree_stop(sig,degrees); 
  if (deg < 0) return 0;
  if (first_stop(&aux,degrees,deg))  
    do {
      num++; 
    } while (next_stop(&aux,degrees));
  return num;
}
